import 'dart:isolate';
import 'dart:ui';
import 'package:flutter/foundation.dart';
import 'package:flutter_downloader/flutter_downloader.dart';
import 'ALDownloaderHandlerInterface.dart';
import 'ALDownloaderPersistentFileManager.dart';
import 'ALDownloaderStatus.dart';

/// ALDownloader
class ALDownloader {
  /// ---------------------- Public API ----------------------

  /// initialize
  ///
  /// called or lazy call
  ///
  /// lazy call：when call [download]
  static initialize() async {
    if (!_isInitial) {
      // a dirty flag that guarantees that this scope is executed only once
      _isInitial = true;

      // FlutterDownloader initialize
      await FlutterDownloader.initialize(debug: false);

      // register FlutterDownloader callback
      FlutterDownloader.registerCallback(_downloadCallback);

      // register the isolate communication service
      _addIsolateNameServerPortService();

      // extract all current tasks from database and execute the tasks that need to be executed
      await _loadAndTryToRunTask();
    }
  }

  /// download
  ///
  /// **parameters**
  ///
  /// [url] url
  ///
  /// [downloaderHandlerInterface] the download handle interface
  static download(String? url,
      {ALDownloaderHandlerInterface? downloaderHandlerInterface}) async {
    if (url == null) throw "ALDownloader | download error = url is null";

    if (downloaderHandlerInterface != null) {
      final aBinder = _ALDownloaderBinder(url, downloaderHandlerInterface);
      _alDownloaderBinders.add(aBinder);
    }

    await initialize();

    final anALDownloadTask = _getALDownloadTaskFromUrl(url);

    if (anALDownloadTask == null) {
      // get the 'physical directory path' and 'file name' of the file by URL
      final alDownloaderPathComponentModel =
          await ALDownloaderPersistentFileManager
              .lazyGetALDownloaderPathModelFromUrl(url);

      // equeued a task
      final taskId = await FlutterDownloader.enqueue(
          url: url,
          savedDir: alDownloaderPathComponentModel.dir,
          fileName: alDownloaderPathComponentModel.fileName,
          showNotification: false,
          openFileFromNotification: false);

      if (taskId != null)
        _addALDownloadTaskOrReplaceALDownloadTaskId(
            url, taskId, DownloadTaskStatus.enqueued);

      debugPrint(
          "ALDownloader | a download task was generated, download status = enqueued，taskId = $taskId");
    } else if (anALDownloadTask.status == DownloadTaskStatus.failed ||
        anALDownloadTask.status == DownloadTaskStatus.canceled) {
      final newTaskIdForRetry =
          await FlutterDownloader.retry(taskId: anALDownloadTask.taskId);
      if (newTaskIdForRetry != null) {
        debugPrint("ALDownloader | newTaskIdForRetry = $newTaskIdForRetry");
        _addALDownloadTaskOrReplaceALDownloadTaskId(
            url, newTaskIdForRetry, DownloadTaskStatus.undefined);
      }
    } else if (anALDownloadTask.status == DownloadTaskStatus.paused) {
      final newTaskIdForResume =
          await FlutterDownloader.resume(taskId: anALDownloadTask.taskId);
      if (newTaskIdForResume != null)
        _addALDownloadTaskOrReplaceALDownloadTaskId(
            url, newTaskIdForResume, DownloadTaskStatus.undefined);
    } else if (anALDownloadTask.status == DownloadTaskStatus.complete) {
      debugPrint(
          "ALDownloader | try to download downloadtaskOfUrl = $url, url is complete");
      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final successHandler = element.downloaderHandlerHolder.successHandler;
          if (successHandler != null) successHandler();
        }
      });
      _alDownloaderBinders.removeWhere((element) => element.forUrl == url);
    } else if (anALDownloadTask.status == DownloadTaskStatus.running) {
      debugPrint(
          "ALDownloader | try to download downloadtaskOfUrl = $url, but url is running");
    } else if (anALDownloadTask.status == DownloadTaskStatus.enqueued) {
      debugPrint(
          "ALDownloader | try to download downloadtaskOfUrl = $url, but url is enqueued");
    } else {
      debugPrint(
          "ALDownloader | try to download downloadtaskOfUrl = $url, but url is undefined");
    }
  }

  /// add a download handle interface
  ///
  /// **parameters**
  ///
  /// [downloaderHandlerInterface] download handle interface
  ///
  /// [url] url
  static void addALDownloaderHandlerInterface(
      ALDownloaderHandlerInterface? downloaderHandlerInterface,
      String? forUrl) {
    if (downloaderHandlerInterface == null || forUrl == null) return;
    final aBinder = _ALDownloaderBinder(forUrl, downloaderHandlerInterface);
    _alDownloaderBinders.add(aBinder);
  }

  /// remove a download handle interface
  ///
  /// **parameters**
  ///
  /// [url] url
  static void removeALDownloaderHandlerInterfaceForUrl(String url) =>
      _alDownloaderBinders.removeWhere((element) => url == element.forUrl);

  /// remove all download handle interfaces
  static void removeALDownloaderHandlerInterfaceForAll() =>
      _alDownloaderBinders.clear();

  /// get the download status of [url]
  ///
  /// **parameters**
  ///
  /// [url] url
  ///
  ///  **return**
  ///
  /// download status [ALDownloaderStatus]
  static ALDownloaderStatus getDownloadStatusForUrl(String url) {
    ALDownloaderStatus alDownloaderStatus;

    try {
      final anALDownloadTask = _getALDownloadTaskFromUrl(url);
      final downloadTaskStatus = anALDownloadTask?.status;

      if (downloadTaskStatus == DownloadTaskStatus.canceled ||
          downloadTaskStatus == DownloadTaskStatus.undefined ||
          downloadTaskStatus == DownloadTaskStatus.enqueued)
        alDownloaderStatus = ALDownloaderStatus.unstarted;
      else if (downloadTaskStatus == DownloadTaskStatus.running)
        alDownloaderStatus = ALDownloaderStatus.downloading;
      else if (downloadTaskStatus == DownloadTaskStatus.paused)
        alDownloaderStatus = ALDownloaderStatus.pausing;
      else if (downloadTaskStatus == DownloadTaskStatus.failed)
        alDownloaderStatus = ALDownloaderStatus.downloadFailed;
      else
        alDownloaderStatus = ALDownloaderStatus.downloadSuccced;
    } catch (error) {
      alDownloaderStatus = ALDownloaderStatus.unstarted;
      debugPrint(
          "ALDownloader | getDownloadStatusForUrl = $url, error = $error");
    }
    return alDownloaderStatus;
  }

  /// get the completed URLS
  ///
  /// urls for resources that download successfully or fail are added to the pool
  ///
  /// **return**
  ///
  /// completed urls - [url: bool]
  static Map<String, bool> get urlResults {
    final Map<String, bool> aMap = {};
    try {
      _alDownloadTasks.forEach((element) {
        if (element.status == DownloadTaskStatus.complete) {
          aMap[element.url] = true;
        } else if (element.status == DownloadTaskStatus.failed) {
          aMap[element.url] = false;
        }
      });
    } catch (error) {
      debugPrint("ALDownloader | get urlResults error = $error");
    }
    return aMap;
  }

  /// cancel download
  ///
  /// the bytes downloaded will be deleted
  ///
  /// **parameters**
  ///
  /// [url] url
  static cancel(String url) async {
    final alDownloadTask = _getALDownloadTaskFromUrl(url);

    if (alDownloadTask == null) return;

    if (alDownloadTask.status == DownloadTaskStatus.running) {
      final taskId = alDownloadTask.taskId;
      await FlutterDownloader.cancel(taskId: taskId);
      alDownloadTask.status = DownloadTaskStatus.canceled;
    }
  }

  /// cancel all tasks that are downloading
  ///
  /// the bytes downloaded will be deleted
  static cancelAll() async {
    await FlutterDownloader.cancelAll();
    for (final downloadTask in _alDownloadTasks) {
      if (downloadTask.status == DownloadTaskStatus.running)
        downloadTask.status = DownloadTaskStatus.canceled;
    }
  }

  /// pause download
  ///
  /// the bytes downloaded will not be deleted
  ///
  /// status of undefined/enqueued/failed will remove the download queue
  ///
  /// **parameters**
  ///
  /// [url] url
  static pause(String url) async {
    try {
      final alDownloadTask = _getALDownloadTaskFromUrl(url);

      if (alDownloadTask == null) {
        debugPrint(
            "ALDownloader | pause, to be suspended url = $url, but url's alDownloadTask is null");
        return;
      }

      debugPrint(
          "ALDownloader | pause, to be suspended url = $url, alDownloadTask.status = ${alDownloadTask.status}");

      final taskId = alDownloadTask.taskId;
      if (alDownloadTask.status == DownloadTaskStatus.undefined ||
          alDownloadTask.status == DownloadTaskStatus.enqueued ||
          alDownloadTask.status == DownloadTaskStatus.failed) {
        _alDownloadTasks.remove(alDownloadTask);
        await FlutterDownloader.remove(
            taskId: taskId, shouldDeleteContent: true);
      } else if (alDownloadTask.status == DownloadTaskStatus.running) {
        await FlutterDownloader.pause(taskId: taskId);
      }
    } catch (error) {
      debugPrint("ALDownloader | pause error = $error");
    }
  }

  /// pause all tasks that are downloading and remove a enqueued task
  static pauseAll() async {
    final aTemp = [];
    aTemp.addAll(_alDownloadTasks);
    for (final alDownloadTask in aTemp) {
      final url = alDownloadTask.url;
      await pause(url);
    }
  }

  /// remove the ALDownloader data corresponding to [url]
  ///
  /// **parameters**
  ///
  /// [url] url
  static remove(String url) async {
    try {
      final alDownloaderTask = _getALDownloadTaskFromUrl(url);

      if (alDownloaderTask == null) {
        debugPrint(
            "ALDownloader | remove alDownloaderTask of url = $url, url's  _ALDownloadTask is null");
        return;
      }

      await FlutterDownloader.remove(
          taskId: alDownloaderTask.taskId,
          shouldDeleteContent:
              true); // Delete a download task from DB & delete file
      _alDownloadTasks.remove(alDownloaderTask);
    } catch (error) {
      debugPrint(
          "ALDownloader | remove alDownloaderTask of url = $url, error = $error");
    }
  }

  /// remove ALDownloader data corresponding to all urls
  static removeAll() async {
    for (final alDownloadTask in _alDownloadTasks) remove(alDownloadTask.url);
  }

  /// ---------------------- Private API ----------------------

  /// maintain custom download tasks
  ///
  /// purpose: avoid frequent I/O
  ///
  /// add a new custom task or update the [url] mapping of the existing task's taskId
  static void _addALDownloadTaskOrReplaceALDownloadTaskId(
      String? url, String taskId, DownloadTaskStatus status) {
    if (url == null) {
      debugPrint(
          "_addALDownloadTaskOrReplaceALDownloadTaskId error url == null");
      return;
    }

    _ALDownloadTask? anALDownloadTask;

    try {
      anALDownloadTask =
          _alDownloadTasks.firstWhere((element) => element.url == url);
      anALDownloadTask.taskId = taskId;
      anALDownloadTask.status = status;
    } catch (error) {
      debugPrint(
          "ALDownloader | _addALDownloadTaskOrReplaceALDownloadTaskId error = $error");
    }

    if (anALDownloadTask == null) {
      anALDownloadTask = _ALDownloadTask(url);
      anALDownloadTask.taskId = taskId;
      anALDownloadTask.status = status;

      _alDownloadTasks.add(anALDownloadTask);
    }
  }

  /// register send port and receive port for [IsolateNameServer]
  ///
  /// used for communication between download isolate and the main isolate
  static void _addIsolateNameServerPortService() {
    IsolateNameServer.registerPortWithName(
        _receivePort.sendPort, _kDownloaderSendPort);
    _receivePort.listen((dynamic data) {
      final id = data[0];
      final status = data[1];
      final progress = data[2];
      _processDataFromPort(id, status, progress);
    });
  }

  /// the callback binded by [FlutterDownloader]
  static void _downloadCallback(
      String id, DownloadTaskStatus status, int progress) {
    final SendPort? send =
        IsolateNameServer.lookupPortByName(_kDownloaderSendPort);
    send?.send([id, status, progress]);
  }

  /// process the [FlutterDownloader]'s callback
  static void _processDataFromPort(
      String id, DownloadTaskStatus status, int progress) {
    final alDownloadTask = _getALDownloadTaskFromTaskId(id);

    if (alDownloadTask == null) {
      debugPrint("not found alDownloadTask， id = $id");
      return;
    }

    final url = alDownloadTask.url;

    _addALDownloadTaskOrReplaceALDownloadTaskId(url, id, status);

    // ignore: non_constant_identifier_names
    final double_progress =
        double.tryParse(((progress / 100).toStringAsFixed(2))) ?? 0;

    if (status == DownloadTaskStatus.complete) {
      debugPrint(
          "ALDownloader | _downloadCallback \n 下载成功 url = $url \nid = $id ");
      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final progressHandler =
              element.downloaderHandlerHolder.progressHandler;
          if (progressHandler != null) progressHandler(double_progress);

          final successHandler = element.downloaderHandlerHolder.successHandler;
          if (successHandler != null) successHandler();
        }
      });
      _alDownloaderBinders.removeWhere((element) => element.forUrl == url);
    } else if (status == DownloadTaskStatus.failed) {
      debugPrint(
          "ALDownloader | _downloadCallback \n download failed url = $url \nid = $id ");
      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final progressHandler =
              element.downloaderHandlerHolder.progressHandler;
          if (progressHandler != null) progressHandler(double_progress);

          final failureHandler = element.downloaderHandlerHolder.failureHandler;
          if (failureHandler != null) failureHandler();
        }
      });
    } else if (status == DownloadTaskStatus.running) {
      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final progressHandler =
              element.downloaderHandlerHolder.progressHandler;
          if (progressHandler != null) progressHandler(double_progress);
        }
      });
    } else if (status == DownloadTaskStatus.paused) {
      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final pausedHandler = element.downloaderHandlerHolder.pausedHandler;
          if (pausedHandler != null) pausedHandler();
        }
      });
    }

    debugPrint(
        "ALDownloader | _downloadCallback \nid = $id \nurl = $url \nstatus = $status \nprogress = $progress \ndouble_progress = $double_progress");
  }

  /// load [FlutterDownloader] local database task to the memory cache, and attempt to execute the tasks
  static _loadAndTryToRunTask() async {
    final tasks = await FlutterDownloader.loadTasks();
    if (tasks != null) {
      tasks.forEach((element) {
        debugPrint(
            "_loadAndTryToRunTask element url = ${element.url}, status = ${element.status}, element url = ${element.taskId}");
      });

      debugPrint(
          "ALDownloader | _loadAndTryToRunTask tasks.length = ${tasks.length}");

      final aList = tasks.map((element) {
        final anALDownloadTask = _ALDownloadTask(element.url);
        anALDownloadTask.taskId = element.taskId;
        anALDownloadTask.status = element.status;
        return anALDownloadTask;
      }).toList();
      _alDownloadTasks.addAll(aList);
    }
  }

  /// by on the [url], get [_ALDownloadTask] from custom download tasks
  static _ALDownloadTask? _getALDownloadTaskFromUrl(String url) {
    _ALDownloadTask? anALDownloadTask;
    try {
      anALDownloadTask =
          _alDownloadTasks.firstWhere((element) => url == element.url);
    } catch (error) {
      debugPrint("ALDownloader | _getALDownloadTaskFromUrl, error = $error");
    }
    return anALDownloadTask;
  }

  /// by on the [url], get task id from custom download tasks
  // ignore: unused_element
  static String? _getTaskIdFromUrl(String url) {
    String? taskId;
    try {
      taskId =
          _alDownloadTasks.firstWhere((element) => url == element.url).taskId;
    } catch (error) {
      debugPrint("ALDownloader | _getTaskIdWith, error = $error");
    }
    return taskId;
  }

  /// by on the [taskId], get task id from custom download tasks
  // ignore: unused_element
  static _ALDownloadTask? _getALDownloadTaskFromTaskId(String taskId) {
    _ALDownloadTask? anALDownloadTask;
    try {
      anALDownloadTask =
          _alDownloadTasks.firstWhere((element) => taskId == element.taskId);
    } catch (error) {
      debugPrint("ALDownloader | _getALDownloadTaskFromTaskId, error = $error");
    }
    return anALDownloadTask;
  }

  /// by the [taskId], get url from custom download tasks
  // ignore: unused_element
  static String? _getUrlWithTaskId(String taskId) {
    String? url;
    try {
      url = _alDownloadTasks
          .firstWhere((element) => taskId == element.taskId)
          .url;
    } catch (error) {
      debugPrint("ALDownloader | _getUrlWithTaskId, error = $error");
    }
    return url;
  }

  /// a dirty flag that guarantees that this scope is executed only once
  static bool _isInitial = false;

  /// custom download tasks
  static final List<_ALDownloadTask> _alDownloadTasks = [];

  /// a binder list for binding element of url and download ininterface
  static final List<_ALDownloaderBinder> _alDownloaderBinders = [];

  /// send port key
  static final _kDownloaderSendPort = "al_downloader_send_port";

  /// receive port
  static final ReceivePort _receivePort = ReceivePort();

  /// privatize constructor
  ALDownloader._();
}

/// class of custom download task
class _ALDownloadTask {
  final String url;

  String taskId = "";
  int progress = 0;
  DownloadTaskStatus status = DownloadTaskStatus.undefined;

  _ALDownloadTask(this.url);
}

/// a binder for binding element of url and download ininterface
/// it may bind more elements in the future
class _ALDownloaderBinder {
  _ALDownloaderBinder(this.forUrl, this.downloaderHandlerHolder);
  final String forUrl;
  final ALDownloaderHandlerInterface downloaderHandlerHolder;
}
